package java.security;

import javax.security.auth.callback.Callback;
import javax.security.auth.callback.PasswordCallback;
import java.io.FileInputStream;
import java.io.File;
import javax.crypto.SecretKey;
import java.util.HashSet;
import java.security.cert.X509Certificate;
import java.util.Collections;
import java.util.Set;
import javax.security.auth.callback.CallbackHandler;
import javax.security.auth.DestroyFailedException;
import java.util.Arrays;
import java.security.spec.AlgorithmParameterSpec;
import javax.security.auth.Destroyable;
import java.io.InputStream;
import java.security.cert.CertificateException;
import java.io.IOException;
import java.io.OutputStream;
import java.util.Enumeration;
import java.util.Date;
import java.security.cert.Certificate;
import sun.security.util.Debug;

public class KeyStore
{
    private static final Debug pdebug;
    private static final boolean skipDebug;
    private static final String KEYSTORE_TYPE = "keystore.type";
    private String type;
    private Provider provider;
    private KeyStoreSpi keyStoreSpi;
    private boolean initialized;
    
    protected KeyStore(KeyStoreSpi keyStoreSpi, Provider provider, String type) {
        this.initialized = false;
        this.keyStoreSpi = keyStoreSpi;
        this.provider = provider;
        this.type = type;
        if (!KeyStore.skipDebug && KeyStore.pdebug != null) {
            KeyStore.pdebug.println("KeyStore." + type.toUpperCase() + " type from: " + this.provider.getName());
        }
    }
    
    public static KeyStore getInstance(String s) throws KeyStoreException {
        try {
            Object[] impl = Security.getImpl(s, "KeyStore", (String)null);
            return new KeyStore((KeyStoreSpi)impl[0], (Provider)impl[1], s);
        }
        catch (NoSuchAlgorithmException ex) {
            throw new KeyStoreException(s + " not found", ex);
        }
        catch (NoSuchProviderException ex2) {
            throw new KeyStoreException(s + " not found", ex2);
        }
    }
    
    public static KeyStore getInstance(String str, String s) throws KeyStoreException, NoSuchProviderException {
        if (s == null || s.length() == 0) {
            throw new IllegalArgumentException("missing provider");
        }
        try {
            Object[] impl = Security.getImpl(str, "KeyStore", s);
            return new KeyStore((KeyStoreSpi)impl[0], (Provider)impl[1], str);
        }
        catch (NoSuchAlgorithmException ex) {
            throw new KeyStoreException(str + " not found", ex);
        }
    }
    
    public static KeyStore getInstance(String str, Provider provider) throws KeyStoreException {
        if (provider == null) {
            throw new IllegalArgumentException("missing provider");
        }
        try {
            Object[] impl = Security.getImpl(str, "KeyStore", provider);
            return new KeyStore((KeyStoreSpi)impl[0], (Provider)impl[1], str);
        }
        catch (NoSuchAlgorithmException ex) {
            throw new KeyStoreException(str + " not found", ex);
        }
    }
    
    public static final String getDefaultType() {
        String s = (String)AccessController.doPrivileged(new PrivilegedAction<String>() {
            @Override
            public String run() {
                return Security.getProperty("keystore.type");
            }
        });
        if (s == null) {
            s = "jks";
            return s;
        }
        return s;
    }
    
    public final Provider getProvider() {
        return this.provider;
    }
    
    public final String getType() {
        return this.type;
    }
    
    public final Key getKey(String s, char[] array) throws KeyStoreException, NoSuchAlgorithmException, UnrecoverableKeyException {
        if (!this.initialized) {
            throw new KeyStoreException("Uninitialized keystore");
        }
        return this.keyStoreSpi.engineGetKey(s, array);
    }
    
    public final Certificate[] getCertificateChain(String s) throws KeyStoreException {
        if (!this.initialized) {
            throw new KeyStoreException("Uninitialized keystore");
        }
        return this.keyStoreSpi.engineGetCertificateChain(s);
    }
    
    public final Certificate getCertificate(String s) throws KeyStoreException {
        if (!this.initialized) {
            throw new KeyStoreException("Uninitialized keystore");
        }
        return this.keyStoreSpi.engineGetCertificate(s);
    }
    
    public final Date getCreationDate(String s) throws KeyStoreException {
        if (!this.initialized) {
            throw new KeyStoreException("Uninitialized keystore");
        }
        return this.keyStoreSpi.engineGetCreationDate(s);
    }
    
    public final void setKeyEntry(String s, Key key, char[] array, Certificate[] array2) throws KeyStoreException {
        if (!this.initialized) {
            throw new KeyStoreException("Uninitialized keystore");
        }
        if (key instanceof PrivateKey && (array2 == null || array2.length == 0)) {
            throw new IllegalArgumentException("Private key must be accompanied by certificate chain");
        }
        this.keyStoreSpi.engineSetKeyEntry(s, key, array, array2);
        return;
    }
    
    public final void setKeyEntry(String s, byte[] array, Certificate[] array2) throws KeyStoreException {
        if (!this.initialized) {
            throw new KeyStoreException("Uninitialized keystore");
        }
        this.keyStoreSpi.engineSetKeyEntry(s, array, array2);
        return;
    }
    
    public final void setCertificateEntry(String s, Certificate certificate) throws KeyStoreException {
        if (!this.initialized) {
            throw new KeyStoreException("Uninitialized keystore");
        }
        this.keyStoreSpi.engineSetCertificateEntry(s, certificate);
        return;
    }
    
    public final void deleteEntry(String s) throws KeyStoreException {
        if (!this.initialized) {
            throw new KeyStoreException("Uninitialized keystore");
        }
        this.keyStoreSpi.engineDeleteEntry(s);
        return;
    }
    
    public final Enumeration<String> aliases() throws KeyStoreException {
        if (!this.initialized) {
            throw new KeyStoreException("Uninitialized keystore");
        }
        return this.keyStoreSpi.engineAliases();
    }
    
    public final boolean containsAlias(String s) throws KeyStoreException {
        if (!this.initialized) {
            throw new KeyStoreException("Uninitialized keystore");
        }
        return this.keyStoreSpi.engineContainsAlias(s);
    }
    
    public final int size() throws KeyStoreException {
        if (!this.initialized) {
            throw new KeyStoreException("Uninitialized keystore");
        }
        return this.keyStoreSpi.engineSize();
    }
    
    public final boolean isKeyEntry(String s) throws KeyStoreException {
        if (!this.initialized) {
            throw new KeyStoreException("Uninitialized keystore");
        }
        return this.keyStoreSpi.engineIsKeyEntry(s);
    }
    
    public final boolean isCertificateEntry(String s) throws KeyStoreException {
        if (!this.initialized) {
            throw new KeyStoreException("Uninitialized keystore");
        }
        return this.keyStoreSpi.engineIsCertificateEntry(s);
    }
    
    public final String getCertificateAlias(Certificate certificate) throws KeyStoreException {
        if (!this.initialized) {
            throw new KeyStoreException("Uninitialized keystore");
        }
        return this.keyStoreSpi.engineGetCertificateAlias(certificate);
    }
    
    public final void store(OutputStream outputStream, char[] array) throws KeyStoreException, IOException, NoSuchAlgorithmException, CertificateException {
        if (!this.initialized) {
            throw new KeyStoreException("Uninitialized keystore");
        }
        this.keyStoreSpi.engineStore(outputStream, array);
        return;
    }
    
    public final void store(KeyStore.LoadStoreParameter loadStoreParameter) throws KeyStoreException, IOException, NoSuchAlgorithmException, CertificateException {
        if (!this.initialized) {
            throw new KeyStoreException("Uninitialized keystore");
        }
        this.keyStoreSpi.engineStore(loadStoreParameter);
        return;
    }
    
    public final void load(InputStream inputStream, char[] array) throws IOException, NoSuchAlgorithmException, CertificateException {
        this.keyStoreSpi.engineLoad(inputStream, array);
        this.initialized = true;
    }
    
    public final void load(KeyStore.LoadStoreParameter loadStoreParameter) throws IOException, NoSuchAlgorithmException, CertificateException {
        this.keyStoreSpi.engineLoad(loadStoreParameter);
        this.initialized = true;
    }
    
    public final KeyStore.Entry getEntry(String s, KeyStore.ProtectionParameter protectionParameter) throws NoSuchAlgorithmException, UnrecoverableEntryException, KeyStoreException {
        if (s == null) {
            throw new NullPointerException("invalid null input");
        }
        if (!this.initialized) {
            throw new KeyStoreException("Uninitialized keystore");
        }
        return this.keyStoreSpi.engineGetEntry(s, protectionParameter);
    }
    
    public final void setEntry(String s, KeyStore.Entry entry, KeyStore.ProtectionParameter protectionParameter) throws KeyStoreException {
        if (s == null || entry == null) {
            throw new NullPointerException("invalid null input");
        }
        if (!this.initialized) {
            throw new KeyStoreException("Uninitialized keystore");
        }
        this.keyStoreSpi.engineSetEntry(s, entry, protectionParameter);
        return;
    }
    
    public final boolean entryInstanceOf(String s, Class<? extends KeyStore.Entry> clazz) throws KeyStoreException {
        if (s == null || clazz == null) {
            throw new NullPointerException("invalid null input");
        }
        if (!this.initialized) {
            throw new KeyStoreException("Uninitialized keystore");
        }
        return this.keyStoreSpi.engineEntryInstanceOf(s, clazz);
    }
    
    static {
        pdebug = Debug.getInstance("provider", "Provider");
        skipDebug = (Debug.isOn("engine=") && !Debug.isOn("keystore"));
    }
    
    public static class PasswordProtection implements KeyStore.ProtectionParameter, Destroyable
    {
        private final char[] password;
        private final String protectionAlgorithm;
        private final AlgorithmParameterSpec protectionParameters;
        private volatile boolean destroyed;
        
        public PasswordProtection(char[] array) {
            this.destroyed = false;
            this.password = ((array == null) ? null : ((char[])array.clone()));
            this.protectionAlgorithm = null;
            this.protectionParameters = null;
        }
        
        public PasswordProtection(char[] array, String protectionAlgorithm, AlgorithmParameterSpec protectionParameters) {
            this.destroyed = false;
            if (protectionAlgorithm == null) {
                throw new NullPointerException("invalid null input");
            }
            this.password = ((array == null) ? null : ((char[])array.clone()));
            this.protectionAlgorithm = protectionAlgorithm;
            this.protectionParameters = protectionParameters;
            return;
        }
        
        public String getProtectionAlgorithm() {
            return this.protectionAlgorithm;
        }
        
        public AlgorithmParameterSpec getProtectionParameters() {
            return this.protectionParameters;
        }
        
        public synchronized char[] getPassword() {
            if (this.destroyed) {
                throw new IllegalStateException("password has been cleared");
            }
            return this.password;
        }
        
        @Override
        public synchronized void destroy() throws DestroyFailedException {
            this.destroyed = true;
            if (this.password != null) {
                Arrays.fill(this.password, ' ');
            }
        }
        
        @Override
        public synchronized boolean isDestroyed() {
            return this.destroyed;
        }
    }
    
    public static class CallbackHandlerProtection implements KeyStore.ProtectionParameter
    {
        private final CallbackHandler handler;
        
        public CallbackHandlerProtection(CallbackHandler handler) {
            if (handler == null) {
                throw new NullPointerException("handler must not be null");
            }
            this.handler = handler;
            return;
        }
        
        public CallbackHandler getCallbackHandler() {
            return this.handler;
        }
    }
    
    public interface Entry
    {
        default public Set<KeyStore.Entry.Attribute> getAttributes() {
            return Collections.emptySet();
        }
        
        public interface Attribute
        {
            public abstract String getName();
            
            public abstract String getValue();
        }
    }
    
    public static final class PrivateKeyEntry implements KeyStore.Entry
    {
        private final PrivateKey privKey;
        private final Certificate[] chain;
        private final Set<KeyStore.Entry.Attribute> attributes;
        
        public PrivateKeyEntry(PrivateKey privateKey, Certificate[] array) {
            this(privateKey, array, Collections.emptySet());
        }
        
        public PrivateKeyEntry(PrivateKey privKey, Certificate[] array, Set<KeyStore.Entry.Attribute> c) {
            if (privKey == null || array == null || c == null) {
                throw new NullPointerException("invalid null input");
            }
            if (array.length == 0) {
                throw new IllegalArgumentException("invalid zero-length input chain");
            }
            Certificate[] chain = (Certificate[])array.clone();
            String type = chain[0].getType();
            int i = 1;
            while (i < chain.length) {
                if (!type.equals(chain[i].getType())) {
                    throw new IllegalArgumentException("chain does not contain certificates of the same type");
                }
                ++i;
            }
            if (!privKey.getAlgorithm().equals(chain[0].getPublicKey().getAlgorithm())) {
                throw new IllegalArgumentException("private key algorithm does not match algorithm of public key in end entity certificate (at index 0)");
            }
            this.privKey = privKey;
            if (chain[0] instanceof X509Certificate && !(chain instanceof X509Certificate[])) {
                System.arraycopy(chain, 0, this.chain = new X509Certificate[chain.length], 0, chain.length);
            }
            else {
                this.chain = chain;
            }
            this.attributes = Collections.unmodifiableSet(new HashSet(c));
            return;
        }
        
        public PrivateKey getPrivateKey() {
            return this.privKey;
        }
        
        public Certificate[] getCertificateChain() {
            return (Certificate[])this.chain.clone();
        }
        
        public Certificate getCertificate() {
            return this.chain[0];
        }
        
        @Override
        public Set<KeyStore.Entry.Attribute> getAttributes() {
            return this.attributes;
        }
        
        @Override
        public String toString() {
            StringBuilder sb = new StringBuilder();
            sb.append("Private key entry and certificate chain with " + this.chain.length + " elements:\r\n");
            Certificate[] chain = this.chain;
            for (int length = chain.length, i = 0; i < length; ++i) {
                sb.append(chain[i]);
                sb.append("\r\n");
            }
            return sb.toString();
        }
    }
    
    public static final class SecretKeyEntry implements KeyStore.Entry
    {
        private final SecretKey sKey;
        private final Set<KeyStore.Entry.Attribute> attributes;
        
        public SecretKeyEntry(SecretKey sKey) {
            if (sKey == null) {
                throw new NullPointerException("invalid null input");
            }
            this.sKey = sKey;
            this.attributes = Collections.emptySet();
            return;
        }
        
        public SecretKeyEntry(SecretKey sKey, Set<KeyStore.Entry.Attribute> c) {
            if (sKey == null || c == null) {
                throw new NullPointerException("invalid null input");
            }
            this.sKey = sKey;
            this.attributes = Collections.unmodifiableSet(new HashSet(c));
            return;
        }
        
        public SecretKey getSecretKey() {
            return this.sKey;
        }
        
        @Override
        public Set<KeyStore.Entry.Attribute> getAttributes() {
            return this.attributes;
        }
        
        @Override
        public String toString() {
            return "Secret key entry with algorithm " + this.sKey.getAlgorithm();
        }
    }
    
    public static final class TrustedCertificateEntry implements KeyStore.Entry
    {
        private final Certificate cert;
        private final Set<KeyStore.Entry.Attribute> attributes;
        
        public TrustedCertificateEntry(Certificate cert) {
            if (cert == null) {
                throw new NullPointerException("invalid null input");
            }
            this.cert = cert;
            this.attributes = Collections.emptySet();
            return;
        }
        
        public TrustedCertificateEntry(Certificate cert, Set<KeyStore.Entry.Attribute> c) {
            if (cert == null || c == null) {
                throw new NullPointerException("invalid null input");
            }
            this.cert = cert;
            this.attributes = Collections.unmodifiableSet(new HashSet(c));
            return;
        }
        
        public Certificate getTrustedCertificate() {
            return this.cert;
        }
        
        @Override
        public Set<KeyStore.Entry.Attribute> getAttributes() {
            return this.attributes;
        }
        
        @Override
        public String toString() {
            return "Trusted certificate entry:\r\n" + this.cert.toString();
        }
    }
    
    public abstract static class Builder
    {
        static final int MAX_CALLBACK_TRIES = 3;
        
        protected Builder() {
        }
        
        public abstract KeyStore getKeyStore() throws KeyStoreException;
        
        public abstract KeyStore.ProtectionParameter getProtectionParameter(String p0) throws KeyStoreException;
        
        public static KeyStore.Builder newInstance(KeyStore keyStore, KeyStore.ProtectionParameter protectionParameter) {
            if (keyStore == null || protectionParameter == null) {
                throw new NullPointerException();
            }
            if (!keyStore.initialized) {
                throw new IllegalArgumentException("KeyStore not initialized");
            }
            return new KeyStore.Builder() {
                private volatile boolean getCalled;
                
                @Override
                public KeyStore getKeyStore() {
                    this.getCalled = true;
                    return keyStore;
                }
                
                @Override
                public KeyStore.ProtectionParameter getProtectionParameter(String s) {
                    if (s == null) {
                        throw new NullPointerException();
                    }
                    if (!this.getCalled) {
                        throw new IllegalStateException("getKeyStore() must be called first");
                    }
                    return protectionParameter;
                }
            };
        }
        
        public static KeyStore.Builder newInstance(String s, Provider provider, File obj, KeyStore.ProtectionParameter protectionParameter) {
            if (s == null || obj == null || protectionParameter == null) {
                throw new NullPointerException();
            }
            if (!(protectionParameter instanceof KeyStore.PasswordProtection) && !(protectionParameter instanceof KeyStore.CallbackHandlerProtection)) {
                throw new IllegalArgumentException("Protection must be PasswordProtection or CallbackHandlerProtection");
            }
            if (!obj.isFile()) {
                throw new IllegalArgumentException("File does not exist or it does not refer to a normal file: " + obj);
            }
            return new KeyStore.Builder.FileBuilder(s, provider, obj, protectionParameter, AccessController.getContext());
        }
        
        public static KeyStore.Builder newInstance(String s, Provider provider, KeyStore.ProtectionParameter protectionParameter) {
            if (s == null || protectionParameter == null) {
                throw new NullPointerException();
            }
            return new KeyStore.Builder() {
                private volatile boolean getCalled;
                private IOException oldException;
                private final PrivilegedExceptionAction<KeyStore> action = new PrivilegedExceptionAction<KeyStore>() {
                    @Override
                    public KeyStore run() throws Exception {
                        KeyStore keyStore;
                        if (provider == null) {
                            keyStore = KeyStore.getInstance(s);
                        }
                        else {
                            keyStore = KeyStore.getInstance(s, provider);
                        }
                        KeyStore.SimpleLoadStoreParameter o = new SimpleLoadStoreParameter(protectionParameter);
                        if (!(protectionParameter instanceof KeyStore.CallbackHandlerProtection)) {
                            keyStore.load(o);
                        }
                        else {
                            int n = 0;
                            while (true) {
                                ++n;
                                try {
                                    keyStore.load(o);
                                    break;
                                }
                                catch (IOException ex) {
                                    if ((ex.getCause() instanceof UnrecoverableKeyException)) {
                                        if (n < 3) {
                                            continue;
                                        }
                                        provider.oldException = ex;
                                    }
                                    throw ex;
                                }
                            }
                        }
                        provider.getCalled = true;
                        return keyStore;
                    }
                };
                
                @Override
                public synchronized KeyStore getKeyStore() throws KeyStoreException {
                    if (this.oldException != null) {
                        throw new KeyStoreException("Previous KeyStore instantiation failed", this.oldException);
                    }
                    try {
                        return (KeyStore)AccessController.doPrivileged(this.action, context);
                    }
                    catch (PrivilegedActionException ex) {
                        throw new KeyStoreException("KeyStore instantiation failed", ex.getCause());
                    }
                }
                
                @Override
                public KeyStore.ProtectionParameter getProtectionParameter(String s) {
                    if (s == null) {
                        throw new NullPointerException();
                    }
                    if (!this.getCalled) {
                        throw new IllegalStateException("getKeyStore() must be called first");
                    }
                    return protectionParameter;
                }
            };
        }
        
        private static final class FileBuilder extends KeyStore.Builder
        {
            private final String type;
            private final Provider provider;
            private final File file;
            private KeyStore.ProtectionParameter protection;
            private KeyStore.ProtectionParameter keyProtection;
            private final AccessControlContext context;
            private KeyStore keyStore;
            private Throwable oldException;
            
            FileBuilder(String type, Provider provider, File file, KeyStore.ProtectionParameter protection, AccessControlContext context) {
                this.type = type;
                this.provider = provider;
                this.file = file;
                this.protection = protection;
                this.context = context;
            }
            
            @Override
            public synchronized KeyStore getKeyStore() throws KeyStoreException {
                if (this.keyStore != null) {
                    return this.keyStore;
                }
                if (this.oldException != null) {
                    throw new KeyStoreException("Previous KeyStore instantiation failed", this.oldException);
                }
                PrivilegedExceptionAction<KeyStore> privilegedExceptionAction = new PrivilegedExceptionAction<KeyStore>() {
                    @Override
                    public KeyStore run() throws Exception {
                        if (!(KeyStore.Builder.FileBuilder.this.protection instanceof KeyStore.CallbackHandlerProtection)) {
                            return this.run0();
                        }
                        int n = 0;
                        while (true) {
                            ++n;
                            try {
                                return this.run0();
                            }
                            catch (IOException ex) {
                                if (n < 3 && ex.getCause() instanceof UnrecoverableKeyException) {
                                    continue;
                                }
                                throw ex;
                            }
                        }
                    }
                    
                    public KeyStore run0() throws Exception {
                        KeyStore keyStore;
                        if (KeyStore.Builder.FileBuilder.this.provider == null) {
                            keyStore = KeyStore.getInstance(KeyStore.Builder.FileBuilder.this.type);
                        }
                        else {
                            keyStore = KeyStore.getInstance(KeyStore.Builder.FileBuilder.this.type, KeyStore.Builder.FileBuilder.this.provider);
                        }
                        Object o = null;
                        try {
                            o = new FileInputStream(KeyStore.Builder.FileBuilder.this.file);
                            char[] array;
                            if ((KeyStore.Builder.FileBuilder.this.protection instanceof KeyStore.PasswordProtection)) {
                                array = ((KeyStore.PasswordProtection)KeyStore.Builder.FileBuilder.this.protection).getPassword();
                                KeyStore.Builder.FileBuilder.this.keyProtection = KeyStore.Builder.FileBuilder.this.protection;
                            }
                            else {
                                CallbackHandler callbackHandler = ((KeyStore.CallbackHandlerProtection)KeyStore.Builder.FileBuilder.this.protection).getCallbackHandler();
                                PasswordCallback o2 = new PasswordCallback("Password for keystore " + KeyStore.Builder.FileBuilder.this.file.getName(), false);
                                callbackHandler.handle(new Callback[] { o2 });
                                array = o2.getPassword();
                                if (array == null) {
                                    throw new KeyStoreException("No password provided");
                                }
                                o2.clearPassword();
                                KeyStore.Builder.FileBuilder.this.keyProtection = new PasswordProtection(array);
                            }
                            keyStore.load(o, array);
                        }
                        finally {
                            if (o != null) {
                                o.close();
                            }
                        }
                    }
                };
                try {
                    return this.keyStore = (KeyStore)AccessController.doPrivileged(privilegedExceptionAction, this.context);
                }
                catch (PrivilegedActionException ex) {
                    this.oldException = ex.getCause();
                    throw new KeyStoreException("KeyStore instantiation failed", this.oldException);
                }
            }
            
            @Override
            public synchronized KeyStore.ProtectionParameter getProtectionParameter(String s) {
                if (s == null) {
                    throw new NullPointerException();
                }
                if (this.keyStore == null) {
                    throw new IllegalStateException("getKeyStore() must be called first");
                }
                return this.keyProtection;
            }
        }
    }
    
    static class SimpleLoadStoreParameter implements KeyStore.LoadStoreParameter
    {
        private final KeyStore.ProtectionParameter protection;
        
        SimpleLoadStoreParameter(KeyStore.ProtectionParameter protection) {
            this.protection = protection;
        }
        
        @Override
        public KeyStore.ProtectionParameter getProtectionParameter() {
            return this.protection;
        }
    }
    
    public interface ProtectionParameter
    {
    }
    
    public interface LoadStoreParameter
    {
        public abstract KeyStore.ProtectionParameter getProtectionParameter();
    }
}
